/// <reference lib="webworker" />

import { expose } from 'comlink'
import Vips from 'wasm-vips'

type Operation<S extends Vips.Image, Params extends unknown[]> = (state: S, ...args: Params) => S
type OperationParameters<T> = T extends Operation<any, infer Params> ? Params : never

const vips = Vips({
  dynamicLibraries: [],
})
let sourceImage: Blob
let vipsImage: Vips.Image
let loaderName: string

function log(message: string, ...args: unknown[]) {
  console.debug(`[vips.worker] ${message}`, ...args)
}

function getImageFormat(loaderName: string) {
  return (
    {
      jpeg: '.jpg',
      png: '.png',
      webp: '.webp',
      avif: '.webp',
      gif: '.webp',
    }[loaderName] ?? '.jpg'
  )
}

async function setImage(source: Blob) {
  log('loading Vips')
  const _vips = await vips
  sourceImage = source
  log('Creating new Vips.Image')
  vipsImage = _vips.Image.newFromBuffer(await source.arrayBuffer()).autorot()
  const vipsLoader = vipsImage.getString('vips-loader')
  const suffixLength = vipsLoader.endsWith('load_buffer') ? 11 : 4
  loaderName = vipsLoader.slice(0, vipsLoader.length - suffixLength)
  log(`Determined image type ${loaderName}`)
}

function flip(image: Vips.Image, axis: 'horizontal' | 'vertical') {
  if (axis === 'horizontal') return image.flipHor()
  else if (axis === 'vertical') return image.flipVer()
  throw new TypeError('Unknown flip axis.')
}

function rotate(image: Vips.Image, angle: number) {
  return image.rotate(angle)
}

export type CropOptions = { left: number; top: number; width: number; height: number }

function crop(image: Vips.Image, options: CropOptions) {
  log(`cropping image`, { ...options })
  return image.crop(options.left, options.top, options.width, options.height)
}

function renderImage(image: Vips.Image, type?: string) {
  const imageFormat = getImageFormat(loaderName)
  log(`rendering ${imageFormat} image`)
  const buffer = image.writeToBuffer(imageFormat).buffer as ArrayBuffer
  return new Blob([buffer], type ? { type } : undefined)
}

function wrapOperation<T extends Operation<Vips.Image, any[]>>(operation: T) {
  return (...args: OperationParameters<T>) => {
    if (!vipsImage) return
    const newImage = operation(vipsImage, ...args)
    vipsImage.delete()
    vipsImage = newImage
  }
}

const workerInterface = {
  setImage,
  renderImage: () => (vipsImage ? renderImage(vipsImage, sourceImage?.type) : null),
  getDimensions: () => ({ width: vipsImage?.width, height: vipsImage?.height }),
  rotate: wrapOperation(rotate),
  flip: wrapOperation(flip),
  crop: wrapOperation(crop),
}

expose(workerInterface)

export type ExposedInterface = typeof workerInterface
