import {
  APIListPaginated,
  APIRemove,
  APIRetrieve,
  APIUpdate,
  createExtendableAPI,
  ExtendableAPI,
} from '@rokoli/bnb/drf'
import { defineStore } from 'pinia'

import { createSteeringURL } from '@/api'
import { steeringAuthInit } from '@/stores/auth'
import { Episode, KeysFrom, TimeSlot } from '@/types'
import { aggregateWithIdsParameter } from '@/util/api'
import { useEpisodeStore } from '@/stores/episodes'

type ReadonlyAttrs = KeysFrom<
  TimeSlot,
  'id' | 'start' | 'end' | 'repetitionOfId' | 'scheduleId' | 'showId'
>
type TimeSlotUpdateData = Omit<TimeSlot, ReadonlyAttrs>
type TimeSlotPartialUpdateData = Partial<TimeSlotUpdateData>

export const useTimeSlotStore = defineStore('timeslots', () => {
  const endpoint = createSteeringURL.prefix('timeslots')
  const { api, base } = createExtendableAPI<TimeSlot>(endpoint, steeringAuthInit)
  const listOperations = APIListPaginated(api)

  useEpisodeUpdateCascadeBehaviour(api)

  return {
    ...base,
    ...listOperations,
    ...APIRetrieve(api, {
      aggregate: aggregateWithIdsParameter(listOperations.listIsolated),
    }),
    ...APIUpdate<TimeSlot, TimeSlotUpdateData, TimeSlotPartialUpdateData>(api),
    ...APIRemove(api),
  }
})

function useEpisodeUpdateCascadeBehaviour(api: ExtendableAPI<TimeSlot>) {
  async function fetchUpdates(episodeIds: (TimeSlot['episodeId'] | undefined)[]) {
    const episodeStore = useEpisodeStore()
    const timeslotStore = useTimeSlotStore()
    const episodeResults = await episodeStore.retrieveMultiple(
      episodeIds.filter((id) => Boolean(id)) as number[],
    )
    await timeslotStore.retrieveMultiple(
      episodeResults
        .filter((result) => result.status === 'fulfilled')
        .flatMap((result) => (result as PromiseFulfilledResult<Episode>).value.timeslotIds),
    )
  }

  api.events.addEventListener('set', (event) => {
    const oldEpisodeId = api.itemMap.value.get(event.detail.id)?.episodeId
    if (oldEpisodeId === undefined) return
    const newEpisodeId = event.detail.object.episodeId
    if (oldEpisodeId !== newEpisodeId)
      setTimeout(() => void fetchUpdates([newEpisodeId, oldEpisodeId]), 0)
  })
}
